(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   Andrew Tolmach
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Omega.
Require Export EqNat.
Require Export Coq.Arith.Div2.
Require Export Even.
Require Import Bool.
Require Import Arith.
Require Export MyTactics.


Set Implicit Arguments.


Lemma lt_dec : forall m n,
  {m < n} + {~m < n}.
(* APT: an even shorter proof *)
Proof.
   intros. 
   destruct (lt_eq_lt_dec m n) as [[LT | EQ] | GT] ; [left | right | right]; omega.
Defined.



Lemma beq_true : forall x,
  beq_nat x x = true.
Proof.
  induction x; simpl; auto.
Qed.

Lemma beq_false : forall x y,
  x <> y ->
  beq_nat x y = false.
Proof.
  induction x; destruct y; simpl; intro NE; auto; elf; auto.
Qed.


Definition eq_nat_dec := eq_nat_dec.


Theorem natInd' : forall P : nat -> Prop,
  P 0 ->
  (forall k, (forall k', k' < k -> P k') -> k <> 0 -> P k) ->
  forall k, P k.
Proof.
  intros P P0 Pk; induction k; auto.
  apply Pk.
  clear IHk; induction k; intros k' LT.

  assert (k' = 0).
  omega.
  subst; auto.

  compare k' 0; intro; subst; auto.
  apply Pk; auto.
  intros; apply IHk.
  omega.
  
  omega.
  
Qed.


Theorem natInd4 : forall P : nat -> Prop,
  P 0 ->
  P 1 ->
  P 2 ->
  P 3 ->
  (forall k, P k -> P (4 + k)) ->
  forall k, P k.
Proof.
  intros P P0 P1 P2 P3 Pk; induction k using natInd'; auto.
  destruct k as [| [| [| [| k]]]]; auto.
  simpl in Pk; apply Pk; apply H; omega.
Qed.


Fixpoint natWordAlign (k:nat) : nat :=
  match k with
    | (0 | 1 | 2 | 3) => 0
    | S (S (S (S k'))) => S (S (S (S (natWordAlign k'))))
  end.


Lemma natWordAlignEq : forall k,
  exists k', natWordAlign k = 4 * k'.
Proof.
  induction k using natInd4; try (exists 0; auto; fail).
  destruct IHk as [k' E]; exists (S k'); simpl; omega.
Qed.


Lemma natWordAlignEq' : forall x,
  exists z, natWordAlign x + z = x /\ z < 4.
Proof.
  induction x using natInd4.
  exists 0; constructor; auto.
  exists 1; constructor; auto.
  exists 2; constructor; auto.
  exists 3; constructor; auto.
  destruct IHx as [z [E LT]]; exists z; constructor; auto.
  simpl; omega.
Qed.


Lemma alignNatWordAlignEq : forall k,
  natWordAlign (4 * k) = 4 * k.
Proof.
  induction k using natInd4; auto.
  replace (4 * (4 + k)) with (16 + 4 * k) by omega.
  generalize dependent (4 * k); clear k; intros.
  simpl; rewrite IHk; auto.
Qed.


Fixpoint blt_nat (k1 k2:nat) {struct k2} : bool :=
  match k2 with
    | 0 => false
    | S k2' =>
      match k1 with
        | 0 => true
        | S k1' => blt_nat k1' k2'
      end
  end.


Lemma natWordAlignAlignEq : forall x,
  natWordAlign (natWordAlign x) = natWordAlign x.
Proof.
  induction x using natInd4; try reflexivity.
  simpl; rewrite IHx; auto.
Qed.


Lemma blt_false : forall y x,
  (x >= y)%nat ->
  blt_nat x y = false.
Proof.
  induction y; intros x GE. auto.
  simpl; destruct x.
  elf; omega.
  apply IHy; omega.
Qed.


Lemma blt_true : forall y x,
  (x < y)%nat ->
  blt_nat x y = true.
Proof.
  induction y; intros x GE.
  elf; omega.
  simpl; destruct x. auto.
  apply IHy; omega.
Qed.
