(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import ProgLog.
Require Import NatUtil.


Set Implicit Arguments.


Fixpoint llist (x:val) (l:list val) : mpred :=
  match l with
    | nil => !(x = 0%word)
    | v::l' =>
      lexists x', x |-> v ** (x + 4)%val |-> x' ** llist x' l'
  end.


Definition wv := Var 4.
Definition vv := Var 5.
Definition tv := Var 6.


Definition reverseBody := (
  wv <- 0;;
  while (vv != 0) (
    tv <- [vv + 4];;
    [vv + 4] <- wv;;
    wv <- vv;;
    vv <- tv
  );;
  ret wv
)%CM.


Notation reverseDef := (fdefi (vv::nil) (wv::tv::nil) 0 reverseBody).

Notation reverseTy := (list val::nil : tslist).

Definition reversePre is i :=
  llist i is.

Definition reversePost is (_ j:val) :=
  llist j (rev is).

Definition P0 : cdhpty := fun _ => None.

Definition inv is (s:cstate) :=
  exists w, exists v,
    (vfEqv (vv :: wv :: tv :: nil) ((wv,w) :: (vv, v) :: nil) (cvfOf s) /\
      (lexists vl, lexists wl, 
        llist v vl ** llist w wl ** !(rev vl ++ wl = rev is)) (cmemOf s)).


Lemma llist_non_null : forall x : val,
  x <> 0 -> forall l,
  mimpl (llist x l)
  (lexists h, lexists t, !(l = h::t) ** 
    lexists x', x |-> h ** (x + 4)%val |-> x' ** llist x' t).
Proof.
  intros x NE l m Hp.
  destruct l as [ | v l].

  simpl in Hp; ssimpl in Hp. elf; auto.

  unfold llist in Hp; fold llist in Hp. ssimpl in Hp.
  ssimpl; eauto.

Qed.


Lemma llist_null : forall l,
  miff (llist 0%word l) (!(l = nil)).
Proof.
  intros l m; constructor; intros Hp; destruct l.

  simpl in Hp; ssimpl in Hp; ssimpl.

  simpl in Hp; ssimpl in Hp.
  (* why doesn't ssimpl in pick up the fact that the cont is on 0? *)
  elf; inversion Hp. inversion H0.

  simpl; ssimplg in Hp.

  ssimpl in Hp. discriminate H.

Qed.


Lemma llist_not_undef : forall l,
  mimpl (llist Vundef l) (!False).
Proof.
  intros l m Hp; destruct l; simpl in Hp; ssimpl in Hp.
  discriminate H.
Qed.


Hint Rewrite llist_non_null using assumption : SLogI.
Hint Rewrite llist_not_undef llist_null : SLogI.


Lemma reverseOk : fdefOk P0 reverseTy 1 reversePre reverseDef reversePost.
Proof.
  fdefBegin.
  unfold reversePre, reverseBody.
  intros vs x m sp Hp vf VFE. 
  ssimpl in Hp. simpl in VFE.
  vcSteps.
  exists (inv vs). split.

  (* base case *)
  econstructor; econstructor; intuition.
  exists vs. exists (@nil val).
  simpl. ssimpl. intuition.

  (* inductive case *)
  clear; intros.
  destruct s'; unfold inv in H; unfold cmemOf, cspOf, cvfOf in * .
  destruct H as [w [v0 [VFE Hp]]]. ssimpl in Hp.
  vcSteps. branchStep.

  (* true branch *)
  sarewriteE in Hp. subst x0.
  vcSteps.
  econstructor; econstructor; intuition.
  exists x3. exists (x2::x1).
  unfold llist; fold llist. ssimpl. searchMatch.
  clear - H. simpl in *. autorewrite with list in H. assumption.

  (* false branch *)
  subst v0. ssimplg in Hp.
  sarewriteE in Hp.
  subst x0. simpl in H. subst x1. assumption.

  (* undefined case *)
  subst v0. sarewriteE in Hp. assumption.

Qed.
