(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Require Import SepSimpl.
Require Import SepSetoid.
Require Import List.

Set Implicit Arguments.


(* sconj lifting *)

Ltac sconjLiftSolver := 
  intros; simple apply sconjComm; simple apply sconjAssocR; sli; smatch0; 
    simple apply sconjComm; assumption.

Lemma sconjLift3Ok : forall A B C D m,
  (A ** B ** C ** D) m ->
  (C ** A ** B ** D) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift4Ok : forall A B C D E m,
  (A ** B ** C ** D ** E) m ->
  (D ** A ** B ** C ** E) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift5Ok : forall A B C D E F m,
  (A ** B ** C ** D ** E ** F) m ->
  (E ** A ** B ** C ** D ** F) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift6Ok : forall A B C D E F G m,
  (A ** B ** C ** D ** E ** F ** G) m ->
  (F ** A ** B ** C ** D ** E ** G) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift7Ok : forall A B C D E F G H m,
  (A ** B ** C ** D ** E ** F ** G ** H) m ->
  (G ** A ** B ** C ** D ** E ** F ** H) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift8Ok : forall A B C D E F G H I m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I) m ->
  (H ** A ** B ** C ** D ** E ** F ** G ** I) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift9Ok : forall A B C D E F G H I J m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m ->
  (I ** A ** B ** C ** D ** E ** F ** G ** H ** J) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift10Ok : forall A B C D E F G H I J K m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m ->
  (J ** A ** B ** C ** D ** E ** F ** G ** H ** I ** K) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift11Ok : forall A B C D E F G H I J K L m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m ->
  (K ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** L) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift12Ok : forall A B C D E F G H I J K L M m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m ->
  (L ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** M) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift13Ok : forall A B C D E F G H I J K L M N m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m ->
  (M ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** N) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift14Ok : forall A B C D E F G H I J K L M N O m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O) m ->
  (N ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** O) m.
Proof.
  sconjLiftSolver.
Qed.

Lemma sconjLift15Ok : forall A B C D E F G H I J K L M N O P m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O ** P) m ->
  (O ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** P) m.
Proof.
  sconjLiftSolver.
Qed.


(* 8.2beta3: workaround for lack of simple apply X in H *)

Ltac sconjCount A :=
  match A with
    | _ ** ?B => 
      let k := sconjCount B in
        constr:(S k)
    | _ => constr:1
  end.

Ltac sconjLiftMatch H k L :=
  let HT := type of H in
  let k' :=
    match HT with
      | (?A ** ?B) _ => sconjCount (A ** B)
      | _ _ => constr:1
    end in
    match eval compute in (blt_nat k k') with
      | true => apply L in H
    end.
    

Ltac sconjLift2 H := sconjLiftMatch H 2 sconjLift2Ok.
Ltac sconjLift3 H := sconjLiftMatch H 3 sconjLift3Ok.
Ltac sconjLift4 H := sconjLiftMatch H 4 sconjLift4Ok.
Ltac sconjLift5 H := sconjLiftMatch H 5 sconjLift5Ok.
Ltac sconjLift6 H := sconjLiftMatch H 6 sconjLift6Ok.
Ltac sconjLift7 H := sconjLiftMatch H 7 sconjLift7Ok.
Ltac sconjLift8 H := sconjLiftMatch H 8 sconjLift8Ok.
Ltac sconjLift9 H := sconjLiftMatch H 9 sconjLift9Ok.
Ltac sconjLift10 H := sconjLiftMatch H 10 sconjLift10Ok.
Ltac sconjLift11 H := sconjLiftMatch H 11 sconjLift11Ok.
Ltac sconjLift12 H := sconjLiftMatch H 12 sconjLift12Ok.
Ltac sconjLift13 H := sconjLiftMatch H 13 sconjLift13Ok.
Ltac sconjLift14 H := sconjLiftMatch H 14 sconjLift14Ok.
Ltac sconjLift15 H := sconjLiftMatch H 15 sconjLift15Ok.


Ltac sconjLiftOfSolver := 
  intros; simple apply sconjComm; sli; auto.

Lemma sconjLift2of2Ok : forall A B m,
  (A ** B) m -> 
  (B ** A) m.
Proof.
  intros A B s Hp; simple apply sconjComm; auto.
Qed.

Lemma sconjLift3of3Ok : forall A B C m,
  (A ** B ** C) m -> 
  (C ** A ** B) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift4of4Ok : forall A B C D m,
  (A ** B ** C ** D) m ->
  (D ** A ** B ** C) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift5of5Ok : forall A B C D E m,
  (A ** B ** C ** D ** E) m ->
  (E ** A ** B ** C ** D) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift6of6Ok : forall A B C D E F m,
  (A ** B ** C ** D ** E ** F) m ->
  (F ** A ** B ** C ** D ** E) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift7of7Ok : forall A B C D E F G m,
  (A ** B ** C ** D ** E ** F ** G) m ->
  (G ** A ** B ** C ** D ** E ** F) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift8of8Ok : forall A B C D E F G H m,
  (A ** B ** C ** D ** E ** F ** G ** H) m ->
  (H ** A ** B ** C ** D ** E ** F ** G) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift9of9Ok : forall A B C D E F G H I m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I) m ->
  (I ** A ** B ** C ** D ** E ** F ** G ** H) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift10of10Ok : forall A B C D E F G H I J m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m ->
  (J ** A ** B ** C ** D ** E ** F ** G ** H ** I) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift11of11Ok : forall A B C D E F G H I J K m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m ->
  (K ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift12of12Ok : forall A B C D E F G H I J K L m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m ->
  (L ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift13of13Ok : forall A B C D E F G H I J K L M m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m ->
  (M ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift14of14Ok : forall A B C D E F G H I J K L M N m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m ->
  (N ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m.
Proof.
  sconjLiftOfSolver.
Qed.

Lemma sconjLift15of15Ok : forall A B C D E F G H I J K L M N O m,
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O) m ->
  (O ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m.
Proof.
  sconjLiftOfSolver.
Qed.

Ltac sconjLift2of2 H := apply sconjLift2of2Ok in H.
Ltac sconjLift3of3 H := apply sconjLift3of3Ok in H.
Ltac sconjLift4of4 H := apply sconjLift4of4Ok in H.
Ltac sconjLift5of5 H := apply sconjLift5of5Ok in H.
Ltac sconjLift6of6 H := apply sconjLift6of6Ok in H.
Ltac sconjLift7of7 H := apply sconjLift7of7Ok in H.
Ltac sconjLift8of8 H := apply sconjLift8of8Ok in H.
Ltac sconjLift9of9 H := apply sconjLift9of9Ok in H.
Ltac sconjLift10of10 H := apply sconjLift10of10Ok in H.
Ltac sconjLift11of11 H := apply sconjLift11of11Ok in H.
Ltac sconjLift12of12 H := apply sconjLift12of12Ok in H.
Ltac sconjLift13of13 H := apply sconjLift13of13Ok in H.
Ltac sconjLift14of14 H := apply sconjLift14of14Ok in H.
Ltac sconjLift15of15 H := apply sconjLift15of15Ok in H.


(* sconj lifting in the goal *)

Ltac gsconjLiftSolver :=
  intros until m; 
    let H := fresh in
      intros H; apply sconjComm in H; apply sconjAssocL in H; sle H;
        smatch0; apply sconjComm; auto.

Lemma gsconjLift2Ok : forall A B C m,
  (B ** A ** C) m ->
  (A ** B ** C) m.
Proof.
  intros; hyr H (sconjAssocL (sconjComm H)); smatch0; apply sconjComm; auto.
Qed.

Lemma gsconjLift3Ok : forall A B C D m,
  (C ** A ** B ** D) m ->
  (A ** B ** C ** D) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift4Ok : forall A B C D E m,
  (D ** A ** B ** C ** E) m ->
  (A ** B ** C ** D ** E) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift5Ok : forall A B C D E F m,
  (E ** A ** B ** C ** D ** F) m ->
  (A ** B ** C ** D ** E ** F) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift6Ok : forall A B C D E F G m,
  (F ** A ** B ** C ** D ** E ** G) m ->
  (A ** B ** C ** D ** E ** F ** G) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift7Ok : forall A B C D E F G H m,
  (G ** A ** B ** C ** D ** E ** F ** H) m ->
  (A ** B ** C ** D ** E ** F ** G ** H) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift8Ok : forall A B C D E F G H I m,
  (H ** A ** B ** C ** D ** E ** F ** G ** I) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift9Ok : forall A B C D E F G H I J m,
  (I ** A ** B ** C ** D ** E ** F ** G ** H ** J) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift10Ok : forall A B C D E F G H I J K m,
  (J ** A ** B ** C ** D ** E ** F ** G ** H ** I ** K) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift11Ok : forall A B C D E F G H I J K L m,
  (K ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** L) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift12Ok : forall A B C D E F G H I J K L M m,
  (L ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** M) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift13Ok : forall A B C D E F G H I J K L M N m,
  (M ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** N) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift14Ok : forall A B C D E F G H I J K L M N O m,
  (N ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** O) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O) m.
Proof.
  gsconjLiftSolver.
Qed.

Lemma gsconjLift15Ok : forall A B C D E F G H I J K L M N O P m,
  (O ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** P) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O ** P) m.
Proof.
  gsconjLiftSolver.
Qed.


Ltac gsconjLift2 := simple apply gsconjLift2Ok.
Ltac gsconjLift3 := simple apply gsconjLift3Ok.
Ltac gsconjLift4 := simple apply gsconjLift4Ok.
Ltac gsconjLift5 := simple apply gsconjLift5Ok.
Ltac gsconjLift6 := simple apply gsconjLift6Ok.
Ltac gsconjLift7 := simple apply gsconjLift7Ok.
Ltac gsconjLift8 := simple apply gsconjLift8Ok.
Ltac gsconjLift9 := simple apply gsconjLift9Ok.
Ltac gsconjLift10 := simple apply gsconjLift10Ok.
Ltac gsconjLift11 := simple apply gsconjLift11Ok.
Ltac gsconjLift12 := simple apply gsconjLift12Ok.
Ltac gsconjLift13 := simple apply gsconjLift13Ok.
Ltac gsconjLift14 := simple apply gsconjLift14Ok.
Ltac gsconjLift15 := simple apply gsconjLift15Ok.


(* sconj lifting at the end of the goal *)
Ltac gsconjLiftOfSolver := 
  intros until m;
    let H := fresh in
      intros H; apply sconjComm in H; sle H; auto.

Lemma gsconjLift2of2Ok : forall A B m,
  (B ** A) m ->
  (A ** B) m.
Proof.
  intros; apply sconjComm; auto.
Qed.

Lemma gsconjLift3of3Ok : forall A B C m,
  (C ** A ** B) m ->
  (A ** B ** C) m. 
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift4of4Ok : forall A B C D m,
  (D ** A ** B ** C) m ->
  (A ** B ** C ** D) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift5of5Ok : forall A B C D E m,
  (E ** A ** B ** C ** D) m ->
  (A ** B ** C ** D ** E) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift6of6Ok : forall A B C D E F m,
  (F ** A ** B ** C ** D ** E) m ->
  (A ** B ** C ** D ** E ** F) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift7of7Ok : forall A B C D E F G m,
  (G ** A ** B ** C ** D ** E ** F) m ->
  (A ** B ** C ** D ** E ** F ** G) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift8of8Ok : forall A B C D E F G H m,
  (H ** A ** B ** C ** D ** E ** F ** G) m ->
  (A ** B ** C ** D ** E ** F ** G ** H) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift9of9Ok : forall A B C D E F G H I m,
  (I ** A ** B ** C ** D ** E ** F ** G ** H) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift10of10Ok : forall A B C D E F G H I J m,
  (J ** A ** B ** C ** D ** E ** F ** G ** H ** I) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift11of11Ok : forall A B C D E F G H I J K m,
  (K ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift12of12Ok : forall A B C D E F G H I J K L m,
  (L ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift13of13Ok : forall A B C D E F G H I J K L M m,
  (M ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift14of14Ok : forall A B C D E F G H I J K L M N m,
  (N ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m.
Proof.
  gsconjLiftOfSolver.
Qed.

Lemma gsconjLift15of15Ok : forall A B C D E F G H I J K L M N O m,
  (O ** A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N) m ->
  (A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ** L ** M ** N ** O) m.
Proof.
  gsconjLiftOfSolver.
Qed.


Ltac gsconjLift2of2 := simple apply gsconjLift2of2Ok.
Ltac gsconjLift3of3 := simple apply gsconjLift3of3Ok.
Ltac gsconjLift4of4 := simple apply gsconjLift4of4Ok.
Ltac gsconjLift5of5 := simple apply gsconjLift5of5Ok.
Ltac gsconjLift6of6 := simple apply gsconjLift6of6Ok.
Ltac gsconjLift7of7 := simple apply gsconjLift7of7Ok.
Ltac gsconjLift8of8 := simple apply gsconjLift8of8Ok.
Ltac gsconjLift9of9 := simple apply gsconjLift9of9Ok.
Ltac gsconjLift10of10 := simple apply gsconjLift10of10Ok.
Ltac gsconjLift11of11 := simple apply gsconjLift11of11Ok.
Ltac gsconjLift12of12 := simple apply gsconjLift12of12Ok.
Ltac gsconjLift13of13 := simple apply gsconjLift13of13Ok.
Ltac gsconjLift14of14 := simple apply gsconjLift14of14Ok.
Ltac gsconjLift15of15 := simple apply gsconjLift15of15Ok.


(* smarter sconj lifting tactics *)

Ltac scl2 H := first [sconjLift2 H | sconjLift2of2 H].
Ltac scl3 H := first [sconjLift3 H | sconjLift3of3 H].
Ltac scl4 H := first [sconjLift4 H | sconjLift4of4 H].
Ltac scl5 H := first [sconjLift5 H | sconjLift5of5 H].
Ltac scl6 H := first [sconjLift6 H | sconjLift6of6 H].
Ltac scl7 H := first [sconjLift7 H | sconjLift7of7 H].
Ltac scl8 H := first [sconjLift8 H | sconjLift8of8 H].
Ltac scl9 H := first [sconjLift9 H | sconjLift9of9 H].
Ltac scl10 H := first [sconjLift10 H | sconjLift10of10 H].
Ltac scl11 H := first [sconjLift11 H | sconjLift11of11 H].
Ltac scl12 H := first [sconjLift12 H | sconjLift12of12 H].
Ltac scl13 H := first [sconjLift13 H | sconjLift13of13 H].
Ltac scl14 H := first [sconjLift14 H | sconjLift14of14 H].
Ltac scl15 H := first [sconjLift15 H | sconjLift15of15 H].

Ltac gscl2 := first [gsconjLift2 | gsconjLift2of2].
Ltac gscl3 := first [gsconjLift3 | gsconjLift3of3].
Ltac gscl4 := first [gsconjLift4 | gsconjLift4of4].
Ltac gscl5 := first [gsconjLift5 | gsconjLift5of5].
Ltac gscl6 := first [gsconjLift6 | gsconjLift6of6].
Ltac gscl7 := first [gsconjLift7 | gsconjLift7of7].
Ltac gscl8 := first [gsconjLift8 | gsconjLift8of8].
Ltac gscl9 := first [gsconjLift9 | gsconjLift9of9].
Ltac gscl10 := first [gsconjLift10 | gsconjLift10of10].
Ltac gscl11 := first [gsconjLift11 | gsconjLift11of11].
Ltac gscl12 := first [gsconjLift12 | gsconjLift12of12].
Ltac gscl13 := first [gsconjLift13 | gsconjLift13of13].
Ltac gscl14 := first [gsconjLift14 | gsconjLift14of14].
Ltac gscl15 := first [gsconjLift15 | gsconjLift15of15].


(* sconj lifting tactics that take positional arguments,
   so they can be used by other tactics *)

Ltac scl k H :=
  match eval compute in k with
    | 0 => idtac
    | 1 => idtac
    | 2 => scl2 H
    | 3 => scl3 H
    | 4 => scl4 H
    | 5 => scl5 H
    | 6 => scl6 H
    | 7 => scl7 H
    | 8 => scl8 H
    | 9 => scl9 H
    | 10 => scl10 H
    | 11 => scl11 H
    | 12 => scl12 H
    | 13 => scl13 H
    | 14 => scl14 H
    | 15 => scl15 H
    | _ => fail "need more cases for sconjLift"
  end.

Ltac gscl k :=
  match eval compute in k with
    | 0 => idtac
    | 1 => idtac
    | 2 => gscl2
    | 3 => gscl3
    | 4 => gscl4
    | 5 => gscl5
    | 6 => gscl6
    | 7 => gscl7
    | 8 => gscl8
    | 9 => gscl9
    | 10 => gscl10
    | 11 => gscl11
    | 12 => gscl12
    | 13 => gscl13
    | 14 => gscl14
    | 15 => gscl15
    | _ => fail "need more cases for gscl"
  end.


(* separation logic reassociator and permuter, 
   copied from work I did for the Flint group  *)

Inductive ntree : Set :=
  | node : ntree -> ntree -> ntree
  | leaf : nat -> ntree.
  (* node could also be a list, though would need custom IH *)

Fixpoint flattenTree (t:ntree) : list nat :=
  match t with
    | node t1 t2 => flattenTree t1 ++ flattenTree t2
    | leaf n => n::nil
  end.


(* Reassociator for the goal *)

(* isLeft is true if this node is not a rightmost child *)
Ltac assocConjG' T isLeft :=
  match eval compute in T with
    | leaf _ => idtac
    | node ?T1 ?T2 => 
      assocConjG' T1 true;
      eapply sconjImplR;
	[ |
	  let h := fresh "m" in
	  let Hp := fresh "Hp" in
	    (intros h Hp; assocConjG' T2 isLeft; simple apply Hp)];
      match eval compute in isLeft with
	| true => simple apply sconjAssocL
	| false => idtac
      end
  end.

Ltac assocConjG T := assocConjG' (T:ntree) false.


(* Reassociator for hypotheses *)

(* this won't work if there are uninstantiated existential variables *)
Ltac assocConj' Hp T isLeft :=
  match eval compute in T with
    | leaf _ => idtac
    | node ?T1 ?T2 => 
      assocConj' Hp T1 true;
      let m1 := fresh "m" in
      let m2 := fresh "m" in
      let J := fresh "J" in
      let E := fresh "E" in
      let Hp1 := fresh "H" in
      let Hp2 := fresh "H" in
	(destruct (sconjE Hp) as [m1 [m2 [J [E [Hp1 Hp2]]]]];
	  assocConj' Hp2 T2 isLeft;
	  hyr Hp (sconjI Hp1 Hp2 J E);
	  clear m1 m2 J E Hp1 Hp2);
      match eval compute in isLeft with
	| true => apply sconjAssocR in Hp
	| false => idtac
      end
  end.

Ltac assocConj Hp T := assocConj' Hp (T:ntree) false.


(* reflective permuter *)

Fixpoint unlist (l:list mpred) : mpred :=
  match l with
    | nil => emp
    | A::l' => A ** unlist l'
  end.


(* kth element of the list, starting from 1 *)
Fixpoint getk (k:nat) (l:list mpred) {struct k} : mpred :=
  match k, l with
    | 1, A::_ => A
    | S k', _::l' => getk k' l'
    | _, _ => emp (* failure *)
  end.

(* list made up of the kth elements of l0, for all k in l  *)
Fixpoint getl (l0:list mpred) (l:list nat) {struct l} : list mpred :=
  match l with
    | nil => nil
    | k::l' => getk k l0 :: getl l0 l'
  end.

(* list with k' elements, starting at k and increasing by 1 *)
Fixpoint baseList (k k':nat) {struct k'} : list nat :=
  match k' with
    | 0 => nil
    | S k'' => k :: baseList (S k) k''
  end.

(* drop the first k elements of l *)
Fixpoint dropList (l:list mpred) (k:nat) {struct k} : list mpred :=
  match k, l with
    | 0, _ | _, nil => l
    | S k', _::l' => dropList l' k'
  end.

Lemma dropListEmpty : forall l,
  dropList l (length l) = nil.
Proof.
  induction l; auto.
Qed.

(* kth element of the list, starting from 0 *)
Fixpoint getk' (k:nat) (l:list mpred) {struct k} : mpred :=
  match k, l with
    | 0, A::_ => A
    | S k', _::l' => getk' k' l'
    | _, _ => emp (* failure *)
  end.

Lemma getk'Eq : forall k l,
  getk (S k) l = getk' k l.
Proof.
  induction k.
  auto.
  simpl.
  destruct l; auto.
  simpl in IHk.
  rewrite IHk; auto.
Qed.


Lemma dropListCons : forall l k,
  length l > k ->
  dropList l k = getk (S k) l :: dropList l (S k).
Proof.
  intros l k; rewrite getk'Eq.
  generalize dependent k.
  induction l.
  
  simpl; intros; elf; omega.
  
  destruct k; simpl.
  auto.
  intros.
  rewrite IHl; auto.
  omega.
Qed.


(* getl for a segment of baseList is the same as dropList *)
Lemma getlOk' : forall l0 k,
  length l0 >= k ->
  miff (unlist (getl l0 (baseList (length l0 - k + 1) k)))
        (unlist (dropList l0 (length l0 - k))).
Proof.
  induction k; intros LT.

  simpl.
  replace (length l0 - 0) with (length l0) by omega.
  rewrite dropListEmpty; intros m; constructor; intro Hp; auto.

  replace (length l0 - S k + 1) with (length l0 - k) by omega.
  simpl.
  rewrite dropListCons.
  2:omega.
  replace (S (length l0 - S k)) with (length l0 - k) by omega.
  simpl.
  rewrite <- IHk by omega.
  replace (length l0 - k + 1) with (S (length l0 - k)) by omega.
  intros m; constructor; auto.

Qed.

  
(* list insertion using blt_nat *)
Fixpoint blist_insert (a:nat) (l:list nat) {struct l} : list nat :=
  match l with
    | nil => (a::nil)
    | a'::l' =>
      if blt_nat a a' then
	a::l
      else
        a'::(blist_insert a l')
  end.

(* duplicate preserving list sort using blt_nat *)
Fixpoint blist_sort (l:list nat) : list nat := 
  match l with
    | nil => nil
    | x::l' => blist_insert x (blist_sort l')
  end.


(* sorting doesn't affect getl *)
Lemma sortGetlOk : forall l0 l,
  miff (unlist (getl l0 l)) (unlist (getl l0 (blist_sort l))).
Proof.
  induction l.
  intros m; constructor; auto.

  simpl; rewrite IHl; clear IHl.
  generalize dependent (blist_sort l); clear l; intros.

  induction l; intros.
  intros m; constructor; auto.
  simpl in * .
  destruct (blt_nat a a0); simpl in * .
  intros m; constructor; auto.
  rewrite <- IHl; clear IHl.
  intros m; constructor; gscl2; auto.
Qed.


(* like unlist, except produces a nicer result *)
Fixpoint unlist' (l:list mpred) : mpred :=
  match l with
    | nil => emp
    | A::nil => A
    | A::l' => A ** unlist' l'
  end.


Lemma unlist'Ok : forall l,
  miff (unlist l) (unlist' l).
Proof.
  induction l; simpl; intros m; constructor; intro Hp; auto.
  destruct l.
  simpl in Hp; sle Hp; auto.
  smatch0; srewrite <- IHl; auto.
  destruct l.
  simpl; sli; auto.
  smatch0; srewrite IHl; auto.
Qed.


(* getl doesn't add or remove anything *)
Lemma getlOk : forall l l0,
  blist_sort l = baseList 1 (length l0) ->
  miff (unlist' (getl l0 l)) (unlist' l0).
Proof.
  intros l l0 E.
  rewrite <- unlist'Ok.
  rewrite <- unlist'Ok.
  replace (unlist l0) with (unlist (dropList l0 (length l0 - length l0))).
  2:replace (length l0 - length l0) with 0 by omega; auto.
  rewrite <- getlOk' by omega.
  replace (length l0 - length l0 + 1) with 1 by omega.
  rewrite <- E.
  rewrite sortGetlOk.
  intros m; constructor; auto.
Qed.


Lemma getlOkG : forall l l0,
  blist_sort l = baseList 1 (length l0) ->
  mimpl (unlist' (getl l0 l)) (unlist' l0).
Proof.
  intros.
  rewrite getlOk; auto.
  intros m Hp; auto.
Qed.


Lemma getlOkH : forall l l0,
  blist_sort l = baseList 1 (length l0) ->
  mimpl (unlist' l0) (unlist' (getl l0 l)).
Proof.
  intros.
  rewrite getlOk; auto.
  intros m Hp; auto.
Qed.


Ltac toList A :=
  match A with
    | ?B ** ?A' => 
      let l := toList A' in
        constr:(B ::l)
    | ?B => constr:(B :: nil)
  end.


(* now some syntax to ease the use of the reassoc tactics *)

Delimit Scope SL_scope with SL.

Coercion leaf : nat >-> ntree. 

Inductive ntlist : Set :=
  | ntnil
  | ntcons : ntree -> ntlist -> ntlist.

Notation ".[ x , .. , y ]" := 
  (ntcons x .. (ntcons y ntnil) ..) : SL_scope.

Fixpoint listTree (l:ntlist) : ntree :=
  match l with
    | ntnil => leaf 0 (* bogus value *)
    | ntcons t ntnil => t
    | ntcons t l' => node t (listTree l')
  end.

Coercion listTree : ntlist >-> ntree.


(* defining the reflexive perm tactics *)

Ltac rpermConjG l0 :=
  let lx := fresh in
  let X := fresh in
  let Y := fresh in
    match goal with
      | |- ?A ?B ?C _ => 
        set (lx := l0);
        vm_compute in lx;
        assert (X := getlOkG lx); subst lx;
        cbv beta delta [blist_sort blist_insert blt_nat] iota in X;
        let l := toList (A B C) in
          assert (Y := X l); clear X;
          cbv beta iota delta [unlist' getl getk length baseList] in Y;
          simple apply (Y (refl_equal _)); clear Y
    end.

Ltac rpermConj Hp l0 :=
  let lx := fresh in
  let X := fresh in
  let Y := fresh in
    match type of Hp with
      | ?A ?B ?C ?m => 
        set (lx := l0);
        vm_compute in lx;
        assert (X := getlOkH lx); subst lx;
        cbv beta delta [blist_sort blist_insert blt_nat] iota in X;
        let l := toList (A B C) in
          assert (Y := X l); clear X;
          cbv beta iota delta [unlist' getl getk length baseList] in Y;
          apply (Y (refl_equal _) m) in Hp; clear Y
    end.


Ltac permConjG T := rpermConjG T.
Ltac permConj Hp T := rpermConj Hp T.


(* combining the two: *)

Tactic Notation "srearr" constr(T) := 
  permConjG (flattenTree (T:ntree));
  assocConjG (T:ntree).

Tactic Notation "srearr" constr(T) "in" hyp(Hp) := 
  permConj Hp (flattenTree (T:ntree));
  assocConj Hp (T:ntree).



