(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export List.
Require Export Val.


(* a little indirection to make expressions less ambiguous *)
Inductive var : Set :=
  | Var : nat -> var.


Module Type VFileSig.

  (* Variable file signature, following the approach for memory of 
     Leroy and Blazy *)

  Parameter vfile : Type.


  (* operations *)

  Parameter base : list var -> vfile.
  Parameter update : vfile -> var -> val -> option vfile.
  Parameter get : vfile -> var -> option val.
  
  Parameter dom : vfile -> list var.


  (* domain *)

  Parameter baseDomEq : forall vars,
    dom (base vars) = vars.

  Parameter updateDomEq : forall vf x v vf',
    update vf x v = Some vf' ->
    dom vf = dom vf'.

  Parameter updateDomE : forall vf x,
    In x (dom vf) -> forall v,
      exists vf', update vf x v = Some vf'.

  Parameter getDomE : forall vf x,
    In x (dom vf) ->
    exists v, get vf x = Some v.


  (* basic operation properties *)
  
  Parameter updateGetEq : forall vf vf' x v,
    update vf x v = Some vf' ->
      get vf' x = Some v.

  Parameter updateGetNeq : forall vf vf' x v,
    update vf x v = Some vf' -> forall y,
      x <> y ->
      get vf' y = get vf y.
 

End VFileSig.
